/*
 * Copyright 2020 ETSI
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holder nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package fr.emse.gitlab.saref.gui;

import java.awt.BorderLayout;
import java.awt.Desktop;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.net.URL;

import javax.swing.ImageIcon;
import javax.swing.JEditorPane;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPasswordField;
import javax.swing.WindowConstants;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import fr.emse.gitlab.saref.SAREF;
import fr.emse.gitlab.saref.SAREFCredentialsProvider;
import fr.emse.gitlab.saref.SAREFPipeline;
import fr.emse.gitlab.saref.SAREFPipeline.Mode;

public class MainFrame extends JFrame implements SAREFCredentialsProvider {

	private static final Logger LOG = LoggerFactory.getLogger(MainFrame.class);

	private File directory = null;
	private SAREFPipeline.Mode mode = Mode.DEVELOP;
	private boolean ignoreExamples = false;
	private boolean ignoreTerms = true;
	private boolean ignoreSite = true;

	private final InfoPanel infoPanel = new InfoPanel();
	private final FunctionalitiesPanel functionalitiesPanel = new FunctionalitiesPanel(this);
	private final ConsolePanel consolePanel = new ConsolePanel();

	public MainFrame() {
		super("SAREF Pipeline");
		setBounds(100, 100, 800, 400);
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		URL url = MainFrame.class.getClassLoader().getResource("etsi.png");
		ImageIcon img = new ImageIcon(url);
		setIconImage(img.getImage());

		getContentPane().add(infoPanel, BorderLayout.NORTH);
		getContentPane().add(consolePanel, BorderLayout.CENTER);
		getContentPane().add(functionalitiesPanel, BorderLayout.SOUTH);

		functionalitiesPanel.addFileChooserActionListener(new FileChooserActionListener(this));
		functionalitiesPanel.addStartActionListener(new StartActionListener(this));

		pack();
		setResizable(true);
		setVisible(true);
		setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
	}

	public void setDirectory(File directory) {
		this.directory = directory;
	}

	public void setMode(SAREFPipeline.Mode mode) {
		this.mode = mode;
	}

	public void setIgnoreSite(boolean ignoreSite) {
		this.ignoreSite = ignoreSite;
	}

	public void setIgnoreExamples(boolean ignoreExamples) {
		this.ignoreExamples = ignoreExamples;
	}

	public void setIgnoreTerms(boolean ignoreTerms) {
		this.ignoreTerms = ignoreTerms;
	}

	private class FileChooserActionListener implements ActionListener {
		private final MainFrame mainFrame;

		private FileChooserActionListener(MainFrame mainFrame) {
			this.mainFrame = mainFrame;
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			JFileChooser fileChooser = new JFileChooser(directory);
			fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
			int option = fileChooser.showOpenDialog(mainFrame);
			if (option == JFileChooser.APPROVE_OPTION) {
				directory = fileChooser.getSelectedFile();
				infoPanel.setDirectory(directory);
				functionalitiesPanel.setEnabled();
			}
		}

	}

	private class StartActionListener implements ActionListener {
		private final MainFrame mainFrame;

		private StartActionListener(MainFrame mainFrame) {
			this.mainFrame = mainFrame;
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			SAREFPipeline pipeline = new SAREFPipeline(mainFrame, directory, mode, ignoreExamples, ignoreTerms,
					ignoreSite);
			consolePanel.clear();
			try {
				boolean reportGenerated = pipeline.run();
				if (reportGenerated) {
					try {
						final File reportHTML = new File(pipeline.targetDir, "site/report.html");
						String link = String.format("<a href=\"%s\">%s</a>", reportHTML.getCanonicalPath(),
								reportHTML.getCanonicalPath());
						String message = SAREF.getMessage("confirm", link);
						JEditorPane messageHTML = new JEditorPane("text/html",
								String.format("<html><body>%s</body></html>", message));
						int n = JOptionPane.showConfirmDialog(mainFrame, messageHTML, SAREF.getMessage("confirm_title"),
								JOptionPane.YES_NO_OPTION);
						if (n == JOptionPane.YES_OPTION) {
							try {
								Desktop.getDesktop().browse(reportHTML.toURI());
							} catch (IOException ex) {
								LOG.error(String.format("Exception while opening the HTML report : ", reportHTML), ex);
							}
						}
					} catch (IOException ex) {
						LOG.warn("Exception", ex);
					}
				}
			} catch (Exception ex) {
				consolePanel.reportError(ex);
				LOG.error("Exception:", ex);
			}

		}

	}

	@Override
	public String getUsername(String hostName) {
		return JOptionPane.showInputDialog(this, SAREF.getMessage(MESSAGE.account_name.toString(), hostName), null);
	}

	@Override
	public char[] getPassword(String hostName) {
		JPanel panel = new JPanel();
		JLabel label = new JLabel(SAREF.getMessage("account_password"));
		JPasswordField pass = new JPasswordField(30);
		panel.add(label);
		panel.add(pass);
		String[] options = new String[] { SAREF.getMessage("OK"), SAREF.getMessage("Cancel") };
		int option = JOptionPane.showOptionDialog(null, panel, "The title", JOptionPane.NO_OPTION,
				JOptionPane.PLAIN_MESSAGE, null, options, options[1]);
		if (option == 0) {
			return pass.getPassword();
		} else {
			LOG.warn(SAREF.getMessage(MESSAGE.account_password_empty.toString(), hostName));
			return "".toCharArray();
		}
	}
}
